import re
from collections import defaultdict
from functools import cache, lru_cache
from pathlib import Path
from pprint import pprint
from typing import List, Dict, Optional

import git
from git import GitCommandError, InvalidGitRepositoryError
from loguru import logger

from aoc_tiles.colors import extension_to_colors
from aoc_tiles.config import Config


class SolutionFinder:
    def __init__(self, config: Config):
        self.config = config
        try:
            self.repository = git.Repo(config.aoc_dir)
        except InvalidGitRepositoryError:
            print(f"{config.aoc_dir} is an InvalidGitRepository. Git functionality will not be available.")
            self.repository = None

        if self.config.session_cookie_path.exists():
            msg = f"This is a security risk!" \
                  f" Ensure that the {self.config.session_cookie_path} "
            if self.git_is_file_tracked(self.config.session_cookie_path):
                msg += "file is not tracked by git!"
                exit(f"[ERROR] Session cookie file is tracked by git. {msg} Aborting!")
            elif not self.git_is_file_ignored(self.config.session_cookie_path):
                msg += " is in your .gitignore!"
                print(f"[WARNING] Session cookie file is not git ignored. {msg}")

    def get_solution_paths_by_year(self, aoc_dir: Path) -> Dict[Optional[int], Dict[int, List[Path]]]:
        day_to_solution_paths = defaultdict(lambda: defaultdict(list))
        year_pattern = re.compile(self.config.year_pattern)
        day_pattern = re.compile(self.config.day_pattern)
        logger.debug("Finding solution files recursively in {}", aoc_dir)
        candidate_paths = sorted(self._find_recursive_solution_files(aoc_dir))
        logger.debug("Candidate paths: {}", candidate_paths)
        for path in candidate_paths:
            years = year_pattern.findall(str(path))
            days = day_pattern.findall(str(path))
            if not years:
                years.append(None)
            for year in years:
                if days:
                    day_to_solution_paths[year][int(days[0])].append(path)

        self._ensure_sorting(day_to_solution_paths)
        # pprint(day_to_solution_paths)
        return day_to_solution_paths

    def _ensure_sorting(self, solution_paths_dict: Dict[int, Dict[int, List[str]]]) -> Dict[int, Dict[int, List[str]]]:
        def sort_key(path: Path):
            suffix = path.suffix.lower()
            if suffix in self.config.language_sorting:
                return self.config.language_sorting.index(suffix), suffix
            return 99999, suffix

        for year in solution_paths_dict:
            for day in solution_paths_dict[year]:
                solution_paths_dict[year][day] = sorted(solution_paths_dict[year][day], key=sort_key)
        return solution_paths_dict

    def _find_recursive_solution_files(self, directory: Path) -> List[Path]:
        if self.config.only_use_solutions_in_git and self.repository is not None:
            files = [Path(s) for s in self.git_get_tracked_files()]
        else:
            files = list(directory.rglob("*"))

        logger.debug("Found {} files", len(files))
        solution_paths = []
        for path in files:
            extension_is_supported = path.suffix in extension_to_colors()
            path_is_excluded = any([path.match(exclude) for exclude in self.config.exclude_patterns])
            if path_is_excluded:
                logger.debug("Excluded: {} because of patterns: {}", path, self.config.exclude_patterns)
            if path.is_file() and extension_is_supported and not path_is_excluded:
                solution_paths.append(path)
        logger.debug("Found {} solution files", len(solution_paths))
        return solution_paths

    def git_is_file_ignored(self, filepath):
        if self.repository is None:
            return False
        try:
            self.repository.git.execute(['git', 'check-ignore', '-q', str(filepath)])
            return True
        except GitCommandError:
            return False

    @lru_cache
    def git_get_tracked_files(self) -> List[str]:
        if self.repository is None:
            return []
        return self.repository.git.ls_files().split('\n')

    def git_is_file_tracked(self, filepath: Path):
        tracked_files = self.git_get_tracked_files()
        return str(filepath) in tracked_files

    def git_add(self, path: Path):
        if repository is not None and path.exists():
            self.repository.git.add(str(path))

    def git_commit_amend(self):
        # Command based on this:
        # https://stackoverflow.com/questions/3284292/can-a-git-hook-automatically-add-files-to-the-commit
        if self.repository is not None:
            self.repository.git.commit('--amend', '-C', 'HEAD', '--no-verify')


def main():
    SolutionFinder(Config()).get_solution_paths_by_year(Path())


if __name__ == "__main__":
    main()



import itertools
import re
import time
from dataclasses import dataclass, fields
from pathlib import Path
from typing import Union, Dict, List, Set

import requests

from aoc_tiles.config import Config

# URL for the personal leaderboard (same for everyone)
PERSONAL_LEADERBOARD_URL = "https://adventofcode.com/{year}/leaderboard/self"


@dataclass
class DayScores:
    time1: Union[str, None] = None
    rank1: Union[str, None] = None
    score1: Union[str, None] = None
    time2: Union[str, None] = None
    rank2: Union[str, None] = None
    score2: Union[str, None] = None


def _parse_leaderboard(leaderboard_path: Path) -> Dict[int, DayScores]:
    no_stars = "You haven't collected any stars... yet."
    start = '<span class="leaderboard-daydesc-both"> *Time *Rank *Score</span>\n'
    end = "</pre>"
    with open(leaderboard_path) as file:
        html = file.read()
        if no_stars in html:
            return {}
        matches = re.findall(rf"{start}(.*?){end}", html, re.DOTALL | re.MULTILINE)
        assert len(matches) == 1, f"Found {'no' if len(matches) == 0 else 'more than one'} leaderboard?!"
        table_rows = matches[0].strip().split("\n")
        day_to_scores = {}
        for line in table_rows:
            day, *scores = re.split(r"\s+", line.strip())
            # replace "-" with None to be able to handle the data later, like if no score existed for the day
            scores = [s if s != "-" else None for s in scores]
            assert len(scores) in (3, 6), f"Number scores for {day=} ({scores}) are not 3 or 6."
            day_to_scores[int(day)] = DayScores(*scores)
        return day_to_scores


def request_leaderboard(year: int, config: Config) -> Dict[int, DayScores]:
    leaderboard_path = config.cache_dir / f"leaderboard{year}.html"
    if leaderboard_path.exists():
        leaderboard = _parse_leaderboard(leaderboard_path)
        less_than_30mins = time.time() - leaderboard_path.lstat().st_mtime < 60 * 30
        if less_than_30mins:
            print(f"Leaderboard for {year} is younger than 30 minutes, skipping download in order to avoid DDOS.")
            return leaderboard
        has_no_none_values = all(itertools.chain(map(fields, leaderboard.values())))
        if has_no_none_values and len(leaderboard) == 25:
            print(f"Leaderboard for {year} is complete, no need to download.")
            return leaderboard

    with open(config.session_cookie_path) as cookie_file:
        session_cookie = cookie_file.read().strip()
        assert len(session_cookie) == 128, f"Session cookie is not 128 characters long, make sure to remove the prefix!"
        data = requests.get(
            PERSONAL_LEADERBOARD_URL.format(year=year),
            headers={"User-Agent": "https://github.com/LiquidFun/aoc_tiles by Brutenis Gliwa"},
            cookies={"session": session_cookie},
        ).text
        leaderboard_path.parent.mkdir(exist_ok=True, parents=True)
        with open(leaderboard_path, "w") as file:
            file.write(data)
    return _parse_leaderboard(leaderboard_path)



import re
from concurrent.futures import ProcessPoolExecutor
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Set, List, Optional, Any

from loguru import logger

from aoc_tiles.colors import extension_to_colors
from aoc_tiles.config import Config
from aoc_tiles.drawer import TileDrawer
from aoc_tiles.html import HTML
from aoc_tiles.leaderboard import DayScores, request_leaderboard
from aoc_tiles.solutions import SolutionFinder


README_TILES_BEGIN = "<!-- AOC TILES BEGIN -->"
README_TILES_END = "<!-- AOC TILES END -->"


@dataclass
class YearData:
    day_to_scores: Dict[int, DayScores]
    day_to_paths: Dict[int, List[Path]]
    day_to_stars: Dict[int, int]


@dataclass
class SolveData:
    year_to_data: Dict[int, YearData]


class TileMaker:
    def __init__(self, config: Config):
        self.config = config
        self.tile_drawer = TileDrawer(config)
        self.solution_finder = SolutionFinder(config)

    def _get_stars(self, solved: DayScores, solution: List[Path]):
        on_leaderboard = 0 if solved is None else bool(solved.rank1) + bool(solved.rank2)
        file_exists = 0 if solution is None else 2
        return {
            "on_leaderboard": on_leaderboard,
            "file_exists": file_exists,
            "either": max(on_leaderboard, file_exists),
            "both": min(on_leaderboard, file_exists),
        }[self.config.count_as_solved_when]

    def compose_solve_data(self) -> SolveData:
        is_solution_paths_needed = self.config.what_to_show_on_right_side in [
            "loc"
        ] or self.config.count_as_solved_when in ["file_exists", "both", "either"]
        solution_paths_by_year = {}
        years = []
        if is_solution_paths_needed:
            solution_paths_by_year = self.solution_finder.get_solution_paths_by_year(self.config.aoc_dir)

            years = solution_paths_by_year.keys()

        if self.config.overwrite_year is not None:
            years = [self.config.overwrite_year]

        is_leaderboard_needed = self.config.what_to_show_on_right_side in [
            "time_and_rank"
        ] or self.config.count_as_solved_when in ["on_leaderboard", "both", "either"]

        solve_data = SolveData({})

        for year in years:
            day_to_solution = solution_paths_by_year.get(year, {})
            day_to_scores = {}
            if is_leaderboard_needed:
                logger.debug("Requesting leaderboard for year {}", year)
                day_to_scores = request_leaderboard(year, self.config)

            day_to_stars = {}

            for day in range(1, 26):
                stars = self._get_stars(day_to_scores.get(day), day_to_solution.get(day))
                day_to_stars[day] = stars

            solve_data.year_to_data[year] = YearData(day_to_scores, day_to_solution, day_to_stars)
        return solve_data

    def handle_day(
        self,
        day: int,
        year: int,
        solutions: List[Path],
        day_scores: Optional[DayScores],
        needs_update: bool,
        stars: int,
    ):
        logger.debug("day={} year={} solutions={}", day, year, solutions)
        languages = []
        for solution in solutions:
            extension = solution.suffix
            if extension in extension_to_colors() and extension not in languages:
                languages.append(extension)
        solution_link = solutions[0] if solutions else None
        day_graphic_path = self.config.image_dir / f"{year:04}/{day:02}.png"
        day_graphic_path.parent.mkdir(parents=True, exist_ok=True)
        if not day_graphic_path.exists() or needs_update:
            self.tile_drawer.draw_tile(f"{day:02}", languages, day_scores, day_graphic_path, stars=stars)
        day_graphic_path = day_graphic_path.relative_to(self.config.aoc_dir)
        return day_graphic_path, solution_link

    def fill_empty_days_in_dict(self, day_to_solutions: Dict[int, List[Path]], max_day) -> None:
        if not self.config.create_all_days and len(day_to_solutions) == 0:
            print(f"Current year has no solutions!")
        for day in range(1, max_day + 1):
            if day not in day_to_solutions:
                day_to_solutions[day] = []

    def handle_year(self, year: int, year_data: YearData):
        print(f"=== Generating table for year {year} ===")
        leaderboard = year_data.day_to_scores
        day_to_solutions = year_data.day_to_paths
        html = HTML()
        with html.tag("h1", align="center"):
            stars = sum(year_data.day_to_stars.values())
            html.push(f"{year} - {stars} ⭐")
        max_solved_day = max(day for day, stars in year_data.day_to_stars.items() if stars > 0)
        max_day = 25 if self.config.create_all_days else max_solved_day
        self.fill_empty_days_in_dict(day_to_solutions, max_day)

        # completed_solutions = dict()
        # completed_cache_path = self.config.cache_dir / f"completed-{year}.json"
        # if completed_cache_path.exists():
        #     with open(completed_cache_path, "r") as file:
        #         completed_solutions = {int(day): solutions for day, solutions in json.load(file).items()}

        day_to_future = {}
        with ProcessPoolExecutor() as executor:
            for day in range(1, max_day + 1):
                solutions = day_to_solutions.get(day, [])
                stars = year_data.day_to_stars[day]
                future = executor.submit(self.handle_day, day, year, solutions, leaderboard.get(day), True, stars=stars)
                day_to_future[day] = future

        for day, future in day_to_future.items():
            tile_path, solution_path = future.result()
            with html.tag("a", href=str(solution_path)):
                html.tag("img", closing=False, src=tile_path.as_posix(), width=self.config.tile_width_px)

        # with open(completed_cache_path, "w") as file:
        #     completed_days = [day for day, scores in leaderboard.items() if scores.time2 is not None]
        #     file.write(
        #         json.dumps({day: solutions for day, solutions in day_to_solutions.items() if day in completed_days})
        #     )

        with open(self.config.readme_path, "r", encoding="utf-8") as file:
            text = file.read()
            begin = README_TILES_BEGIN
            end = README_TILES_END
            assert begin in text and end in text, (
                f"Could not find AOC TILES markers '{begin}' and '{end}' in the "
                f"README.md! Make sure to add them to the README at {self.config.readme_path}."
            )
            pattern = re.compile(rf"{begin}.*{end}", re.DOTALL | re.MULTILINE)
            new_text = pattern.sub(f"{begin}\n{html}\n{end}", text)

        with open(self.config.readme_path, "w", encoding="utf-8") as file:
            file.write(str(new_text))

    def _ensure_is_not_running_already(self):
        if self.config.aoc_tiles_dir.exists():
            if self.config.running_lock_path in self.config.aoc_tiles_dir.iterdir():
                print("AoC-Tiles is already running! Remove running.lock if this is not the case.")
                exit()

    def make_tiles(self):
        self._ensure_is_not_running_already()
        print("Running AoC-Tiles")
        solve_data = self.compose_solve_data()
        logger.info("Found {} years with solutions", len(solve_data.year_to_data))
        for year, data in sorted(solve_data.year_to_data.items(), reverse=True):
            logger.debug("year={} data={}", year, data)
            self.handle_year(year, data)

        if self.config.auto_add_tiles_to_git in ["add", "amend"]:
            self.solution_finder.git_add(self.config.image_dir)
            self.solution_finder.git_add(self.config.readme_path)

        if self.config.auto_add_tiles_to_git in ["amend"]:
            with open(self.config.running_lock_path, "w") as file:
                file.write("")
            try:
                self.solution_finder.git_commit_amend()
            finally:
                print("Could not amend commit. Maybe there is nothing to amend?")
                self.config.running_lock_path.unlink()


def main():
    TileMaker(Config()).make_tiles()



import sys
from dataclasses import dataclass, field
from pathlib import Path
from typing import Literal, List, Tuple, Optional, Union

from PIL import ImageColor
from loguru import logger


@dataclass
class Config:
    aoc_dir: Union[str, Path] = field(default="./", metadata={"help": "Path to the AoC directory.", "type": str})
    readme_path: Union[str, Path] = field(init=False)
    session_cookie_path: Union[str, Path] = field(init=False)
    aoc_tiles_dir: Union[str, Path] = field(init=False)
    image_dir: Union[str, Path] = field(init=False)
    cache_dir: Union[str, Path] = field(init=False)

    verbose: bool = field(default=False, metadata={"help": "Whether to print debug information."})

    what_to_show_on_right_side: Literal["auto", "checkmark", "time_and_rank", "loc"] = field(
        default="auto", metadata={
            "help": "What information to display on the right side of each tile. "
                    "'checkmark' only displays a checkmark for each part if the day is solved. "
                    "'time_and_rank' displays the time and rank on the global leaderboard (requires session.cookie). "
                    "'loc' displays the number of lines of code of the solution (not implemented. "
                    "'auto' will use 'time_and_rank' if session.cookie exists, otherwise 'checkmark'."}
    )
    count_as_solved_when: Literal["auto", "on_leaderboard", "file_exists", "either", "both"] = field(
        default="auto",
        metadata={
            "help": "Condition to count a task as solved. Note that 'on_leaderboard', 'either' and 'both' require a "
                    "session cookie. 'auto' will use 'both' if session.cookie exists, otherwise 'file_exists'."
        },
    )
    language_sorting: List[str] = field(
        default_factory=list,
        metadata={
            "help": "Preferred language extensions order for sorting. For example 'py,rs,js' will make Python "
                    "solutions appear first, then Rust, then JavaScript, then everything else (alphabetically)."
        },
    )
    create_all_days: bool = field(default=False, metadata={"help": "Whether to create entries for all days upfront."})

    auto_add_tiles_to_git: Literal["no", "add", "amend"] = field(default="no", metadata={
        "help": "Whether to automatically add the tile images to git. 'add' will add new files, 'amend' will add "
                "and amend the commit with the new files. 'no' will not add the files to git."})
    only_use_solutions_in_git: bool = field(default=True, metadata={
        "help": "If true, only solutions will be considered which are tracked by git (git added), "
                "otherwise all solutions will be used. This is useful for example to ignore auto-generated"
                "files, like '.d' in Rust or '.o' files in C++."})

    year_pattern: str = field(
        default=r"(?<!\d)(20[123]\d)(?!\d)",
        metadata={
            "help": "Regex pattern for matching years. This extracts the first group as the year and parses it as an "
                    "integer. Make sure that other numbers are not matched by this pattern! For example, "
                    "using negative lookbehind and lookaheads is encouraged to avoid matching longer numbers!"
        },
    )
    day_pattern: str = field(
        default=r"(?<!\d)([012]?\d)(?!\d)", metadata={"help": "Regex pattern for matching days. Same as year_pattern."}
    )
    exclude_patterns: List[str] = field(
        default_factory=list, metadata={
            "help": "A list of comma separated glob patterns to ignore when looking for solutions. "
                    "Listing the paths works too. "
                    "For example: '*.py,*.js', '2023/05/05.c' or '2021/**.py'."
                    "Make sure to escape the patterns with single quotes when running from the shell and with"
                    "double quotes when using in the yaml to avoid shell expansion!"
        }
    )

    overwrite_year: Optional[int] = field(
        default=None,
        metadata={
            "help": "If your repository only contains a single year and it cannot be parsed from the path, then you "
                    "should use this to overwrite the year. Every solution is presumed to be for this year."
        },
    )

    contrast_improvement_type: Literal["none", "outline", "dark"] = field(
        default="outline",
        metadata={
            "help": "Some languages have very light colors and are hard to see with a white font. Here you can choose "
                    "how the text color changes when the background is too light. 'dark' makes the font dark, "
                    "'outline' adds a black outline."
        },
    )
    contrast_improvement_threshold: int = field(
        default=30, metadata={"help": "Threshold for contrast improvement feature (between 0 and 255)."}
    )
    outline_color: Union[str, Tuple] = field(
        default="#6C6A6A", metadata={"help": "Color used for outlining elements.", "type": str}
    )
    not_completed_color: Union[str, Tuple] = field(
        default="#333333", metadata={"help": "Color to signify incomplete tasks.", "type": str}
    )
    top100_color: Union[str, Tuple] = field(
        default="#ffdd00", metadata={"help": "Color to highlight top 100 ranking. Only used if session"
                                             "cookie is provided.", "type": str}
    )
    text_color: Union[str, Tuple] = field(default="#FFFFFF", metadata={"help": "Text color.", "type": str})

    tile_width_px: str = field(default="161px", metadata={"help": "Width of tiles in pixels. You likely don't need"
                                                                  "to change this."})

    def __post_init__(self):
        self.aoc_dir = Path(self.aoc_dir)

        if not hasattr(self, "readme_path"):
            readmes = [path for path in self.aoc_dir.iterdir() if path.name.lower() == "readme.md"]
            if len(readmes) == 0:
                exit(f"[ERROR] No README.md found in the root directory of the repository '{self.aoc_dir}'.")
            elif len(readmes) > 1:
                exit(f"[ERROR] Multiple README.md files found in the root directory of the repository {readmes}.")
            self.readme_path = readmes[0]

        if not hasattr(self, "aoc_tiles_dir"):
            self.aoc_tiles_dir = self.aoc_dir / ".aoc_tiles"
            self.aoc_tiles_dir.mkdir(exist_ok=True)

        self.running_lock_path = self.aoc_tiles_dir / "running.lock"

        if not hasattr(self, "session_cookie_path"):
            self.session_cookie_path = self.aoc_tiles_dir / "session.cookie"
            if not self.session_cookie_path.exists():
                self.session_cookie_path = self.aoc_dir / "session.cookie"

        if not hasattr(self, "image_dir"):
            self.image_dir = self.aoc_tiles_dir / "tiles"

        if not hasattr(self, "cache_dir"):
            self.cache_dir = self.aoc_tiles_dir / "cache"

        if self.count_as_solved_when == "auto":
            self.count_as_solved_when = "both" if self.session_cookie_path.exists() else "file_exists"

        if self.what_to_show_on_right_side == "auto":
            self.what_to_show_on_right_side = "time_and_rank" if self.session_cookie_path.exists() else "checkmark"

        self.outline_color = ImageColor.getrgb(self.outline_color)
        self.not_completed_color = ImageColor.getrgb(self.not_completed_color)
        self.text_color = ImageColor.getrgb(self.text_color)
        self.top100_color = ImageColor.getrgb(self.top100_color)

        for i, suffix in enumerate(self.language_sorting):
            if not suffix.startswith("."):
                self.language_sorting[i] = "." + suffix

        logger.remove()
        if self.verbose:
            logger.add(sys.stderr, level="DEBUG")

        logger.debug(self)



import math
from functools import partial
from pathlib import Path
from typing import List, Tuple, Union, Dict

from PIL import ImageColor, Image
from PIL.ImageDraw import ImageDraw

from aoc_tiles.colors import color_similarity, darker_color, extension_to_colors
from aoc_tiles.config import Config
from aoc_tiles.fonts import main_font, secondary_font
from aoc_tiles.leaderboard import DayScores


def format_time(time: str) -> str:
    """Formats time as mm:ss if the time is below 1 hour, otherwise it returns >1h to a max of >24h

    >>> format_time("00:58:32")
    '58:32'
    >>> format_time(">1h")
    '  >1h'
    """
    time = time.replace("&gt;", ">")
    if ">" in time:
        formatted = time
    else:
        h, m, s = time.split(":")
        formatted = f">{h}h" if int(h) >= 1 else f"{m:02}:{s:02}"
    return f"{formatted:>5}"


class TileDrawer:
    def __init__(self, config: Config):
        self.config = config

    def draw_tile(
            self, day: str, languages: List[str], day_scores: Union[DayScores, None], path: Path, stars: int
    ):
        """Saves a graphic for a given day and year. Returns the path to it."""
        image = self.get_alternating_background(languages, stars == 2)
        drawer = ImageDraw(image)
        text_kwargs = {"fill": self.config.text_color}

        # Get all colors of the day, check if any one is similar to TEXT_COLOR
        # If yes, add outline
        for language in languages:
            color = ImageColor.getrgb(extension_to_colors()[language])
            if color_similarity(color, self.config.text_color, self.config.contrast_improvement_threshold):
                if "outline" in self.config.contrast_improvement_type:
                    text_kwargs["stroke_width"] = 1
                    text_kwargs["stroke_fill"] = self.config.outline_color
                if "dark" in self.config.contrast_improvement_type:
                    text_kwargs["fill"] = self.config.not_completed_color
                break

        draw_text = lambda *args, **kwargs: drawer.text(*args, **kwargs, **text_kwargs)
        draw_line = partial(drawer.line, fill=text_kwargs["fill"], width=2)

        # === Left side ===
        draw_text((3, -5), "Day", align="left", font=main_font(20))
        draw_text((1, -10), str(day), align="center", font=main_font(75))

        # Calculate font size based on number of characters, because it might overflow
        lang_as_str = " ".join(languages)
        lang_font_size = max(6, int(18 - max(0, len(lang_as_str) - 8) * 1.3))
        draw_text((0, 74), lang_as_str, align="left", font=secondary_font(lang_font_size))

        # === Right side (P1 & P2) ===
        for part in (1, 2):
            y = 50 if part == 2 else 0
            time = getattr(day_scores, f"time{part}", None)
            rank = getattr(day_scores, f"rank{part}", None)

            color_override = self.config.top100_color if rank and int(rank) <= 100 else self.config.text_color
            text_kwargs["fill"] = color_override

            if stars >= part:
                draw_text((104, -5 + y), f"P{part} ", align="left", font=main_font(25))

                if self.config.what_to_show_on_right_side == "checkmark":
                    draw_line((160, 35 + y, 150, 25 + y))
                    draw_line((160, 35 + y, 180, 15 + y))

                if self.config.what_to_show_on_right_side == "time_and_rank":
                    draw_text((105, 25 + y), "time", align="right", font=secondary_font(10))
                    draw_text((105, 35 + y), "rank", align="right", font=secondary_font(10))
                    draw_text((143, 3 + y), format_time(time), align="right", font=secondary_font(18))
                    draw_text((133, 23 + y), f"{rank:>6}", align="right", font=secondary_font(18))

                if self.config.what_to_show_on_right_side == "loc":
                    raise NotImplementedError("loc is not implemented yet")

            else:
                # Draw cross
                draw_line((140, 15 + y, 160, 35 + y))
                draw_line((140, 35 + y, 160, 15 + y))

        if day_scores is None and not languages:
            draw_line((15, 85, 85, 85))

        # === Divider lines ===
        draw_line((100, 5, 100, 95), width=1)
        draw_line((105, 50, 195, 50), width=1)

        image.save(path)

    def get_alternating_background(self, languages, both_parts_completed=True, *, stripe_width=20):
        colors = [ImageColor.getrgb(extension_to_colors()[language]) for language in languages]
        if len(colors) == 1:
            colors.append(darker_color(colors[0]))
        image = Image.new("RGB", (200, 100), self.config.not_completed_color)

        def fill_with_colors(colors, fill_only_half):
            for x in range(image.width):
                for y in range(image.height):
                    if fill_only_half and x / image.width + y / image.height > 1:
                        continue
                    image.load()[x, y] = colors[((x + y) // stripe_width) % len(colors)]

        fill_with_colors([self.config.not_completed_color, darker_color(self.config.not_completed_color)], False)
        if colors:
            fill_with_colors(colors, not both_parts_completed)
        return image

    def draw_star(self, drawer: ImageDraw, at: Tuple[int, int], size=9, color="#ffff0022", num_points=5):
        """Draws a star at the given position"""
        diff = math.pi * 2 / num_points / 2
        points: List[Tuple[float, float]] = []
        for angle in [diff * i - math.pi / 2 for i in range(num_points * 2)]:
            factor = size if len(points) % 2 == 0 else size * 0.4
            points.append((at[0] + math.cos(angle) * factor, at[1] + math.sin(angle) * factor))
        drawer.polygon(points, fill=color)



from functools import cache
from pathlib import Path

from PIL import ImageFont

FONTS_PATH = Path(__file__).parent / "resources" / "fonts"


@cache
def get_font(size: int, path: str):
    return ImageFont.truetype(str(path), size)


# Note that the fonts sizes are specifically adjusted to the following fonts, if you change the fonts
# you might need to adjust the font sizes and text locations in the rest of the script.


def main_font(size: int) -> ImageFont:
    return get_font(size, FONTS_PATH / "PaytoneOne.ttf")


def secondary_font(size: int) -> ImageFont:
    return get_font(size, FONTS_PATH / "SourceCodePro-Regular.otf")



class HTMLTag:
    """A convenience class for creating HTML tags, can be used in a with statement, which will close the tag."""

    def __init__(self, parent: "HTML", tag: str, closing: bool = True, **kwargs):
        self.parent = parent
        self.tag = tag
        self.closing = closing
        self.kwargs = kwargs
        attributes = "".join(f' {k}="{v}"' for k, v in self.kwargs.items())
        self.parent.push(f"<{self.tag}{attributes}>", depth=self.closing)

    def __enter__(self):
        pass

    def __exit__(self, *args):
        if self.closing:
            self.parent.push(f"</{self.tag}>", depth=-self.closing)


class HTML:
    """A convenience class to create HTML code, in order be able to close the tags and add the necessary syntax.

    Originally, this was used a lot more, however at this point it is only needed for a single with statement.
    """

    tags: list[str] = []
    depth = 0

    def push(self, tag: str, depth=0):
        if depth < 0:
            self.depth += depth
        self.tags.append("  " * self.depth + tag)
        if depth > 0:
            self.depth += depth

    def tag(self, tag: str, closing: bool = True, **kwargs):
        return HTMLTag(self, tag, closing, **kwargs)

    def __str__(self):
        return "\n".join(self.tags)



import argparse
import dataclasses
import functools
from dataclasses import fields
from pathlib import Path
from typing import Literal, Union, get_args, List, get_origin

import rich.traceback
from loguru import logger

from aoc_tiles.config import Config
from aoc_tiles.make_tiles import TileMaker


def literal_or_error(value, literal_type: type):
    type_literals = get_args(literal_type)
    if value not in type_literals:
        raise argparse.ArgumentTypeError(f"Expected {literal_type}, but received '{value}'")
    return value


def type_for_field(field):
    if "type" in field.metadata:
        return field.metadata["type"]
    elif getattr(field.type, "__origin__", None) is Literal:
        return functools.partial(literal_or_error, literal_type=field.type)
    elif getattr(field.type, "__origin__", None) is Path:
        return Path
    elif get_origin(field.type) == list or get_origin(field.type) == List:
        return lambda x: x.split(",")
    elif field.type == bool:

        def bool_parser(value):
            if value.lower() in {"true", "1", "yes", "y"}:
                return True
            elif value.lower() in {"false", "0", "no", "n"}:
                return False
            else:
                raise argparse.ArgumentTypeError(f"Invalid value for boolean: {value}")

        return bool_parser
    return field.type


def cli_parse_config(datacls):
    parser = argparse.ArgumentParser(description="CLI for Config dataclass")
    for field in fields(datacls):
        if field.init:
            if_possible_values = ""
            if getattr(field.type, "__origin__", None) is Literal:
                if_possible_values = f"\nPossible values: [{','.join(field.type.__args__)}]"
            if_default = f'\nDefault: "{field.default}"' if field.default != dataclasses.MISSING else ""
            kwargs = {
                "type": type_for_field(field),
                "help": field.metadata.get("help", field.name) + if_possible_values + if_default,
            }

            if hasattr(field, "default") and field.default != dataclasses.MISSING:
                kwargs["default"] = field.default
            elif hasattr(field, "default_factory") and field.default_factory != dataclasses.MISSING:
                kwargs["default"] = field.default_factory()

            if "default" in kwargs and isinstance(kwargs["default"], list):
                kwargs["action"] = "store"

            if field.type == bool:
                del kwargs["type"]
                kwargs["action"] = "store_true"
            # print(kwargs)
            parser.add_argument(f'--{field.name.replace("_", "-")}', **kwargs)

    parser.add_argument(
        "positional args are ignored",
        nargs="*",
        help="Any non-keyword arguments are ignored. This is because pre-commit passes all arguments to the hook,"
             " and we don't want to fail the hook because of that.",
    )

    args = vars(parser.parse_args())
    del args["positional args are ignored"]
    return datacls(**args)


def main():
    rich.traceback.install()
    config = cli_parse_config(Config)
    TileMaker(config).make_tiles()


if __name__ == "__main__":
    main()



# Location of yaml file where file extensions are mapped to colors
from functools import lru_cache
from pathlib import Path
from typing import Tuple, Dict

import yaml
from loguru import logger

GITHUB_LANGUAGES_PATH = Path(__file__).parent / "resources" / "github_languages.yaml"

excludes = ["GCC Machine Description"]
includes = {".ipynb": "#DA5B0B"}


@lru_cache
def extension_to_colors() -> Dict[str, str]:
    extension_to_color = {}
    with open(GITHUB_LANGUAGES_PATH) as file:
        logger.debug("Loading github_languages.yaml from {}", GITHUB_LANGUAGES_PATH)
        yaml_loader = yaml.CLoader if yaml.__with_libyaml__ else yaml.Loader
        if not yaml.__with_libyaml__:
            logger.warning("Using slow yaml parser (0.5s vs 0.1s)!")
        github_languages = yaml.load(file, Loader=yaml_loader)
        logger.debug("Loaded github_languages.yaml from {}", GITHUB_LANGUAGES_PATH)
        for language, data in github_languages.items():
            if "color" in data and "extensions" in data and data["type"] == "programming" and language not in excludes:
                for extension in data["extensions"]:
                    extension_to_color[extension.lower()] = data["color"]

    extension_to_color.update(includes)

    return extension_to_color


def darker_color(c: Tuple[int, int, int, int]) -> Tuple[int, int, int, int]:
    return c[0] - 10, c[1] - 10, c[2] - 10, 255


# Luminance of color
def luminance(color):
    return 0.299 * color[0] + 0.587 * color[1] + 0.114 * color[2]


# How similar is color_a to color_b
def color_similarity(color_a, color_b, threshold):
    return abs(luminance(color_a) - luminance(color_b)) < threshold
